from __future__ import annotations

import json
import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

import boto3
import pytest
from typer.testing import CliRunner

from aws_annoying.main import app

from ._helpers import PRINTENV_PY, normalize_console_output, repeat_options

if TYPE_CHECKING:
    from pytest_snapshot.plugin import Snapshot

    from tests.conftest import Invoker

# * Command `load-variables` cannot use Typer CLI runner because it uses `os.execvpe` internally,
# * which replaces the current process with the new one, breaking pytest runtime.
# * But tests that does not reach the `os.execvpe` statement can use Typer CLI runner (or provide `--no-replace` flag).
runner = CliRunner()

pytestmark = [
    pytest.mark.integration,
    pytest.mark.usefixtures("use_localstack"),
]


@pytest.fixture
def set_terminal_width() -> int:
    """Use small terminal to simplify test assertions (due to secret random suffix)."""
    return 80


def setup_resources(*, env_base: dict[str, str] | None = None) -> dict[str, Any]:
    """Set up AWS resources."""
    _variables: dict[Literal["secrets", "parameters"], dict[str, Any]] = {
        "secrets": {
            # Pass to CLI arguments
            "my-app/django-sensitive-settings": {
                "DJANGO_SECRET_KEY": "my-secret-key",
            },
        },
        "parameters": {
            # Pass to CLI arguments
            "/my-app/django-settings": {
                "DJANGO_SETTINGS_MODULE": "config.settings.local",
                "DJANGO_ALLOWED_HOSTS": "*",
                "DJANGO_DEBUG": "False",
            },
            # Pass to execution environment
            "/my-app/override": {
                "DJANGO_ALLOWED_HOSTS": "127.0.0.1,192.168.0.2",
            },
        },
    }

    # Secrets
    secretsmanager = boto3.client("secretsmanager")
    secrets = {}
    for name, value in _variables["secrets"].items():
        secret = secretsmanager.create_secret(
            Name=name,
            SecretString=json.dumps(value),
        )
        secrets[name] = {"data": value, "resource": secret}

    # Parameters
    ssm = boto3.client("ssm")
    parameters = {}
    for name, value in _variables["parameters"].items():
        ssm.put_parameter(
            Name=name,
            Value=json.dumps(value),
            Type="String",
        )
        parameter = ssm.get_parameter(Name=name)["Parameter"]
        parameters[name] = {"data": value, "resource": parameter}

    env = env_base or os.environ | {
        # Direct environment variables
        "LOAD_AWS_CONFIG__900_override": parameters["/my-app/override"]["resource"]["ARN"],
        "DJANGO_SETTINGS_MODULE": "config.settings.development",
    }
    load_resources = [
        secrets["my-app/django-sensitive-settings"]["resource"]["ARN"],
        parameters["/my-app/django-settings"]["resource"]["ARN"],
    ]

    return {
        # Test environment
        "env": env,
        "load_resources": load_resources,
        # Resources
        "secrets": {k: v["resource"]["ARN"] for k, v in secrets.items()},
        "parameters": {k: v["resource"]["ARN"] for k, v in parameters.items()},
    }


printenv_py = str(PRINTENV_PY.relative_to(Path.cwd()))
printenv = [printenv_py, "DJANGO_SETTINGS_MODULE", "DJANGO_SECRET_KEY", "DJANGO_DEBUG", "DJANGO_ALLOWED_HOSTS"]


def test_nothing(snapshot: Snapshot) -> None:
    """If nothing is provided, the command should do nothing."""
    # Arrange
    # ...

    # Act
    result = runner.invoke(
        app,
        [
            "load-variables",
        ],
    )

    # Assert
    assert result.exit_code == 0
    snapshot.assert_match(normalize_console_output(result.stdout), "stdout.txt")


def test_unsupported_resource(snapshot: Snapshot) -> None:
    """If unsupported resource ARN provided, should exit with error."""
    # Arrange
    # ...

    # Act
    result = runner.invoke(
        app,
        [
            "load-variables",
            "--arns",
            "arn:aws:s3:::my-bucket/my-object",
            printenv_py,
        ],
    )

    # Assert
    assert result.exit_code == 1
    snapshot.assert_match(normalize_console_output(result.stdout), "stdout.txt")


def test_basic(snapshot: Snapshot, invoke_cli: Invoker) -> None:
    """Test basic usage."""
    # Arrange
    setup = setup_resources()

    # Act
    result = invoke_cli(
        "load-variables",
        *repeat_options("--arns", setup["load_resources"]),
        "--no-replace",
        "--",
        *printenv,
        env=setup["env"],
    )

    # Assert
    assert result.returncode == 0
    snapshot.assert_match(normalize_console_output(result.stdout), "stdout.txt")
    assert result.stderr == ""


def test_replace_quiet(snapshot: Snapshot, invoke_cli: Invoker) -> None:
    """Test the most common practical use-case."""
    # Arrange
    setup = setup_resources()

    # Act
    result = invoke_cli(
        "load-variables",
        *repeat_options("--arns", setup["load_resources"]),
        "--env-prefix",
        "LOAD_AWS_CONFIG__",
        "--quiet",
        "--",
        *printenv,
        env=setup["env"],
    )

    # Assert
    assert result.returncode == 0
    snapshot.assert_match(normalize_console_output(result.stdout), "stdout.txt")
    assert result.stderr == ""


def test_env_prefix(snapshot: Snapshot, invoke_cli: Invoker) -> None:
    """Test prefixed environment variables support."""
    # Arrange
    setup = setup_resources()

    # Act
    result = invoke_cli(
        "load-variables",
        *repeat_options("--arns", setup["load_resources"]),
        "--env-prefix",
        "LOAD_AWS_CONFIG__",
        "--no-replace",
        "--",
        *printenv,
        env=setup["env"],
    )

    # Assert
    assert result.returncode == 0
    snapshot.assert_match(normalize_console_output(result.stdout), "stdout.txt")
    assert result.stderr == ""


def test_dry_run(snapshot: Snapshot, invoke_cli: Invoker) -> None:
    """If dry-run mode enabled, it shouldn't load variables."""
    # Arrange
    setup = setup_resources()

    # Act
    result = invoke_cli(
        "load-variables",
        *repeat_options("--arns", setup["load_resources"]),
        "--env-prefix",
        "LOAD_AWS_CONFIG__",
        "--no-replace",
        "--dry-run",
        "--",
        *printenv,
        env=setup["env"],
    )

    # Assert
    assert result.returncode == 0
    snapshot.assert_match(normalize_console_output(result.stdout), "stdout.txt")
    assert result.stderr == ""


def test_overwrite_env(snapshot: Snapshot, invoke_cli: Invoker) -> None:
    """Test `--overwrite-env` flag. If provided, it should overwrite the existing environment variables."""
    # Arrange
    setup = setup_resources()

    # Act
    result = invoke_cli(
        "load-variables",
        *repeat_options("--arns", setup["load_resources"]),
        "--env-prefix",
        "LOAD_AWS_CONFIG__",
        "--no-replace",
        "--overwrite-env",
        "--",
        *printenv,
        env=setup["env"],
    )

    # Assert
    assert result.returncode == 0
    snapshot.assert_match(normalize_console_output(result.stdout), "stdout.txt")
    assert result.stderr == ""


@pytest.mark.parametrize(
    argnames="arn",
    argvalues=[
        # TODO(lasuillard): Moto Secrets Manager does not respond with error for non-existing resources
        # "arn:aws:secretsmanager:us-east-1:123456789012:secret:unknown-secret", # noqa: ERA001;
        "arn:aws:ssm:us-east-1:123456789012:parameter/unknown-parameter",
    ],
    ids=[
        # "secretsmanager",
        "ssm",
    ],
)
def test_resource_not_found(snapshot: Snapshot, invoke_cli: Invoker, arn: str) -> None:
    """Test with resource does not exists."""
    # Arrange
    setup = setup_resources()

    # Act
    result = invoke_cli(
        "load-variables",
        *repeat_options("--arns", setup["load_resources"]),
        "--arns",
        arn,
        "--no-replace",
        "--",
        *printenv,
        env=setup["env"],
    )

    # Assert
    assert result.returncode == 1
    snapshot.assert_match(normalize_console_output(result.stdout), "stdout.txt")
    assert result.stderr == ""



from __future__ import annotations

import sys
from contextlib import contextmanager
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Generator


# https://stackoverflow.com/questions/38634988/check-if-program-runs-in-debug-mode
def is_debugger_active() -> bool:  # pragma: no cover
    """Return if debugger is active."""
    return hasattr(sys, "gettrace") and sys.gettrace() is not None


@contextmanager
def input_as_args() -> Generator[None, None, None]:  # pragma: no cover
    """Context manager modifying `sys.argv` to pass CLI arguments via input while using debugger."""
    if is_debugger_active():
        args = input("Arguments: ")
        if args:
            sys.argv.extend(args.split(" "))

    yield



from __future__ import annotations

import boto3
import typer
from rich import print  # noqa: A004

from .app import app


@app.command()
def ecs_task_definition_lifecycle(
    *,
    family: str = typer.Option(
        ...,
        help="The name of the task definition family.",
        show_default=False,
    ),
    keep_latest: int = typer.Option(
        ...,
        help="Number of latest (revision) task definitions to keep.",
        show_default=False,
        min=1,
        max=100,
    ),
    dry_run: bool = typer.Option(
        False,  # noqa: FBT003
        help="Do not perform any changes, only show what would be done.",
    ),
) -> None:
    """Execute ECS task definition lifecycle."""
    if dry_run:
        print("⚠️ Dry run mode enabled. Will not perform any actual changes.")

    ecs = boto3.client("ecs")

    # Get all task definitions for the family
    response_iter = ecs.get_paginator("list_task_definitions").paginate(
        familyPrefix=family,
        status="ACTIVE",
        sort="ASC",
    )
    task_definition_arns = []
    for response in response_iter:
        task_definition_arns.extend(response["taskDefinitionArns"])

    # Sort by revision number
    task_definition_arns.sort(key=lambda arn: int(arn.split(":")[-1]))

    # Keep the latest N task definitions
    expired_taskdef_arns = task_definition_arns[:-keep_latest]
    for arn in expired_taskdef_arns:
        if not dry_run:
            ecs.deregister_task_definition(taskDefinition=arn)

        # ARN like: "arn:aws:ecs:<region>:<account-id>:task-definition/<family>:<revision>"
        _, family_revision = arn.split(":task-definition/")
        print(f"✅ Deregistered task definition [yellow]{family_revision!r}[/yellow]")



# flake8: noqa: F401
from __future__ import annotations

import aws_annoying.ecs_task_definition_lifecycle
import aws_annoying.load_variables
import aws_annoying.mfa
from aws_annoying.utils.debugger import input_as_args

# App with all commands registered
from .app import app

__all__ = ("app",)


def entrypoint() -> None:  # pragma: no cover
    app()


if __name__ == "__main__":  # pragma: no cover
    with input_as_args():
        entrypoint()



from pathlib import Path

from .command_builder import repeat_options
from .string_ import normalize_console_output

__all__ = ("PRINTENV_PY", "normalize_console_output", "printenv_py", "repeat_options")

PRINTENV_PY = (Path(__file__).parent / "scripts" / "printenv.py").absolute()



# flake8: noqa: B008
from __future__ import annotations

import json
import os
import subprocess
from typing import Any, NoReturn, Optional

import boto3
import typer
from rich.console import Console
from rich.table import Table

from .app import app


@app.command(
    context_settings={
        # Allow extra arguments for user provided command
        "allow_extra_args": True,
        "ignore_unknown_options": True,
    },
)
def load_variables(  # noqa: PLR0913
    *,
    ctx: typer.Context,
    arns: list[str] = typer.Option(
        [],
        metavar="ARN",
        help=(
            "ARNs of the secret or parameter to load."
            " The variables are loaded in the order of the ARNs,"
            " overwriting the variables with the same name in the order of the ARNs."
        ),
    ),
    env_prefix: Optional[str] = typer.Option(
        None,
        help="Prefix of the environment variables to load the ARNs from.",
        show_default=False,
    ),
    overwrite_env: bool = typer.Option(
        False,  # noqa: FBT003
        help="Overwrite the existing environment variables with the same name.",
    ),
    quiet: bool = typer.Option(
        False,  # noqa: FBT003
        help="Suppress all outputs from this command.",
    ),
    dry_run: bool = typer.Option(
        False,  # noqa: FBT003
        help="Print the progress only. Neither load variables nor run the command.",
    ),
    replace: bool = typer.Option(
        True,  # noqa: FBT003
        help=(
            "Replace the current process (`os.execvpe`) with the command."
            " If disabled, run the command as a `subprocess`."
        ),
    ),
) -> NoReturn:
    """Wrapper command to run command with variables from AWS resources injected as environment variables.

    This script is intended to be used in the ECS environment, where currently AWS does not support
    injecting whole JSON dictionary of secrets or parameters as environment variables directly.

    It first loads the variables from the AWS sources then runs the command with the variables injected as environment variables.

    In addition to `--arns` option, you can provide ARNs as the environment variables by providing `--env-prefix`.
    For example, if you have the following environment variables:

    ```shell
    export LOAD_AWS_CONFIG__001_app_config=arn:aws:secretsmanager:...
    export LOAD_AWS_CONFIG__002_db_config=arn:aws:ssm:...
    ```

    You can run the following command:

    ```shell
    aws-annoying load-variables --env-prefix LOAD_AWS_CONFIG__ -- ...
    ```

    The variables are loaded in the order of option provided, overwriting the variables with the same name in the order of the ARNs.
    Existing environment variables are preserved by default, unless `--overwrite-env` is provided.
    """  # noqa: E501
    console = Console(quiet=quiet, emoji=False)

    command = ctx.args
    if not command:
        console.print("⚠️ No command provided. Exiting...")
        raise typer.Exit(0)

    # Mapping of the ARNs by index (index used for ordering)
    map_arns_by_index = {str(idx): arn for idx, arn in enumerate(arns)}
    if env_prefix:
        console.print(f"🔍 Loading ARNs from environment variables with prefix: {env_prefix!r}")
        arns_env = {
            key.removeprefix(env_prefix): value for key, value in os.environ.items() if key.startswith(env_prefix)
        }
        console.print(f"🔍 Found {len(arns_env)} sources from environment variables.")
        map_arns_by_index = arns_env | map_arns_by_index

    # Briefly show the ARNs
    table = Table("Index", "ARN")
    for idx, arn in sorted(map_arns_by_index.items()):
        table.add_row(idx, arn)

    console.print(table)

    # Retrieve the variables
    loader = VariableLoader(dry_run=dry_run, console=console)
    try:
        variables = loader.load(map_arns_by_index)
    except Exception as exc:  # noqa: BLE001
        console.print(f"❌ Failed to load the variables: {exc!s}")
        raise typer.Exit(1) from None

    # Prepare the environment variables
    env = os.environ.copy()
    if overwrite_env:
        env.update(variables)
    else:
        # Update variables, preserving the existing ones
        for key, value in variables.items():
            env.setdefault(key, str(value))

    # Run the command with the variables injected as environment variables, replacing current process
    console.print(f"🚀 Running the command: [bold orchid]{' '.join(command)}[/bold orchid]")
    if replace:  # pragma: no cover (not coverable)
        os.execvpe(command[0], command, env=env)  # noqa: S606
        # The above line should never return

    result = subprocess.run(command, env=env, check=False)  # noqa: S603
    raise typer.Exit(result.returncode)


# Type aliases for readability
_ARN = str
_Variables = dict[str, Any]


class VariableLoader:  # noqa: D101
    def __init__(self, *, console: Console | None = None, dry_run: bool) -> None:
        """Initialize the VariableLoader.

        Args:
            dry_run: Whether to run in dry-run mode.
            console: Rich console instance.
        """
        self.console = console or Console(quiet=True)
        self.dry_run = dry_run

    # TODO(lasuillard): Currently not using pagination (do we need more than 10-20 secrets or parameters each?)
    #                   ; consider adding it if needed
    def load(self, map_arns: dict[str, _ARN]) -> dict[str, Any]:
        """Load the variables from the AWS Secrets Manager and SSM Parameter Store.

        Each secret or parameter should be a valid dictionary, where the keys are the variable names
        and the values are the variable values.

        The items are merged in the order of the key of provided mapping, overwriting the variables with the same name
        in the order of the keys.
        """
        self.console.print("🔍 Retrieving variables from AWS resources...")
        if self.dry_run:
            self.console.print("⚠️ Dry run mode enabled. Variables won't be loaded from AWS.")

        # Split the ARNs by resource types
        secrets_map, parameters_map = {}, {}
        for idx, arn in map_arns.items():
            if arn.startswith("arn:aws:secretsmanager:"):
                secrets_map[idx] = arn
            elif arn.startswith("arn:aws:ssm:"):
                parameters_map[idx] = arn
            else:
                msg = f"Unsupported resource: {arn!r}"
                raise ValueError(msg)

        # Retrieve variables from AWS resources
        secrets: dict[str, _Variables]
        parameters: dict[str, _Variables]
        if self.dry_run:
            secrets = {idx: {} for idx, _ in secrets_map.items()}
            parameters = {idx: {} for idx, _ in parameters_map.items()}
        else:
            secrets = self._retrieve_secrets(secrets_map)
            parameters = self._retrieve_parameters(parameters_map)

        self.console.print(f"✅ Retrieved {len(secrets)} secrets and {len(parameters)} parameters.")

        # Merge the variables in order
        full_variables = secrets | parameters  # Keys MUST NOT conflict
        merged_in_order = {}
        for _, variables in sorted(full_variables.items()):
            merged_in_order.update(variables)

        return merged_in_order

    def _retrieve_secrets(self, secrets_map: dict[str, _ARN]) -> dict[str, _Variables]:
        """Retrieve the secrets from AWS Secrets Manager."""
        if not secrets_map:
            return {}

        secretsmanager = boto3.client("secretsmanager")

        # Retrieve the secrets
        arns = list(secrets_map.values())
        response = secretsmanager.batch_get_secret_value(SecretIdList=arns)
        if errors := response["Errors"]:
            msg = f"Failed to retrieve secrets: {errors!r}"
            raise ValueError(msg)

        # Parse the secrets
        secrets = response["SecretValues"]
        result = {}
        for secret in secrets:
            arn = secret["ARN"]
            order_key = next(key for key, value in secrets_map.items() if value == arn)
            data = json.loads(secret["SecretString"])
            if not isinstance(data, dict):
                msg = f"Secret data must be a valid dictionary, but got: {type(data)!r}"
                raise TypeError(msg)

            result[order_key] = data

        return result

    def _retrieve_parameters(self, parameters_map: dict[str, _ARN]) -> dict[str, _Variables]:
        """Retrieve the parameters from AWS SSM Parameter Store."""
        if not parameters_map:
            return {}

        ssm = boto3.client("ssm")

        # Retrieve the parameters
        parameter_names = list(parameters_map.values())
        response = ssm.get_parameters(Names=parameter_names, WithDecryption=True)
        if errors := response["InvalidParameters"]:
            msg = f"Failed to retrieve parameters: {errors!r}"
            raise ValueError(msg)

        # Parse the parameters
        parameters = response["Parameters"]
        result = {}
        for parameter in parameters:
            arn = parameter["ARN"]
            order_key = next(key for key, value in parameters_map.items() if value == arn)
            data = json.loads(parameter["Value"])
            if not isinstance(data, dict):
                msg = f"Parameter data must be a valid dictionary, but got: {type(data)!r}"
                raise TypeError(msg)

            result[order_key] = data

        return result



from __future__ import annotations

import configparser
from pathlib import Path  # noqa: TC003
from typing import Optional

import boto3
import typer
from pydantic import BaseModel, ConfigDict
from rich import print  # noqa: A004
from rich.prompt import Prompt

from ._app import mfa_app

_CONFIG_INI_SECTION = "aws-annoying:mfa"


@mfa_app.command()
def configure(  # noqa: PLR0913
    *,
    mfa_profile: Optional[str] = typer.Option(
        None,
        help="The MFA profile to configure.",
    ),
    mfa_source_profile: Optional[str] = typer.Option(
        None,
        help="The AWS profile to use to retrieve MFA credentials.",
    ),
    mfa_serial_number: Optional[str] = typer.Option(
        None,
        help="The MFA device serial number. It is required if not persisted in configuration.",
        show_default=False,
    ),
    mfa_token_code: Optional[str] = typer.Option(
        None,
        help="The MFA token code.",
        show_default=False,
    ),
    aws_credentials: Path = typer.Option(  # noqa: B008
        "~/.aws/credentials",
        help="The path to the AWS credentials file.",
    ),
    aws_config: Path = typer.Option(  # noqa: B008
        "~/.aws/config",
        help="The path to the AWS config file. Used to persist the MFA configuration.",
    ),
    persist: bool = typer.Option(
        True,  # noqa: FBT003
        help="Persist the MFA configuration.",
    ),
) -> None:
    """Configure AWS profile for MFA."""
    # Expand user home directory
    aws_credentials = aws_credentials.expanduser()
    aws_config = aws_config.expanduser()

    # Load configuration
    mfa_config, exists = _MfaConfig.from_ini_file(aws_config, _CONFIG_INI_SECTION)
    if exists:
        print(f"⚙️ Loaded MFA configuration from AWS config ({aws_config}).")

    mfa_profile = (
        mfa_profile
        or mfa_config.mfa_profile
        # _
        or Prompt.ask("👤 Enter name of MFA profile to configure", default="mfa")
    )
    mfa_source_profile = (
        mfa_source_profile
        or mfa_config.mfa_source_profile
        or Prompt.ask("👤 Enter AWS profile to use to retrieve MFA credentials", default="default")
    )
    mfa_serial_number = (
        mfa_serial_number
        or mfa_config.mfa_serial_number
        # _
        or Prompt.ask("🔒 Enter MFA serial number")
    )
    mfa_token_code = (
        mfa_token_code
        # _
        or Prompt.ask("🔑 Enter MFA token code")
    )

    # Get credentials
    print(f"💬 Retrieving MFA credentials using profile [bold]{mfa_source_profile}[/bold]")
    session = boto3.session.Session(profile_name=mfa_source_profile)
    sts = session.client("sts")
    response = sts.get_session_token(
        SerialNumber=mfa_serial_number,
        TokenCode=mfa_token_code,
    )
    credentials = response["Credentials"]

    # Update MFA profile in AWS credentials
    print(f"✅ Updating MFA profile ([bold]{mfa_profile}[/bold]) to AWS credentials ({aws_credentials})")
    _update_credentials(
        aws_credentials,
        mfa_profile,  # type: ignore[arg-type]
        access_key=credentials["AccessKeyId"],
        secret_key=credentials["SecretAccessKey"],
        session_token=credentials["SessionToken"],
    )

    # Persist MFA configuration
    if persist:
        print(
            f"✅ Persisting MFA configuration in AWS config ({aws_config}),"
            f" in [bold]{_CONFIG_INI_SECTION}[/bold] section.",
        )
        mfa_config.mfa_profile = mfa_profile
        mfa_config.mfa_source_profile = mfa_source_profile
        mfa_config.mfa_serial_number = mfa_serial_number
        mfa_config.save_ini_file(aws_config, _CONFIG_INI_SECTION)
    else:
        print("⚠️ MFA configuration not persisted.")


class _MfaConfig(BaseModel):
    model_config = ConfigDict(extra="ignore")

    mfa_profile: Optional[str] = None
    mfa_source_profile: Optional[str] = None
    mfa_serial_number: Optional[str] = None

    def save_ini_file(self, path: Path, section_key: str) -> None:
        """Save configuration to an AWS config file."""
        config_ini = configparser.ConfigParser()
        config_ini.read(path)
        config_ini.setdefault(section_key, {})
        for k, v in self.model_dump(exclude_none=True).items():
            config_ini[section_key][k] = v

        with path.open("w") as f:
            config_ini.write(f)

    @classmethod
    def from_ini_file(cls, path: Path, section_key: str) -> tuple[_MfaConfig, bool]:
        """Load configuration from an AWS config file, with boolean indicating if the config already exists."""
        config_ini = configparser.ConfigParser()
        config_ini.read(path)
        if config_ini.has_section(section_key):
            section = dict(config_ini.items(section_key))
            return cls.model_validate(section), True

        return cls(), False


def _update_credentials(path: Path, profile: str, *, access_key: str, secret_key: str, session_token: str) -> None:
    credentials_ini = configparser.ConfigParser()
    credentials_ini.read(path)
    credentials_ini.setdefault(profile, {})
    credentials_ini[profile]["aws_access_key_id"] = access_key
    credentials_ini[profile]["aws_secret_access_key"] = secret_key
    credentials_ini[profile]["aws_session_token"] = session_token
    with path.open("w") as f:
        credentials_ini.write(f)



import re


def normalize_console_output(output: str) -> str:
    """Normalize the console output for easier comparison."""
    # Remove leading and trailing spaces
    output = output.strip()

    # Unwrap each line
    output = re.sub(r"[ ]+\n", " ", output)

    return output  # noqa: RET504



from __future__ import annotations

import typer

app = typer.Typer(
    pretty_exceptions_short=True,
    pretty_exceptions_show_locals=False,
    rich_markup_mode="rich",
    no_args_is_help=True,
)
